import { h, getAssetPath } from "@stencil/core";
import { HubUrl } from '../../utils/common';
import Iqons from '@nimiq/iqons/dist/iqons.min.js';
import HubApi from '@nimiq/hub-api';
import { Hexagon, Close } from '../../utils/icons';
var Status;
(function (Status) {
    Status[Status["Idle"] = 0] = "Idle";
    Status[Status["Progress"] = 1] = "Progress";
    Status[Status["Success"] = 2] = "Success";
    Status[Status["Error"] = 3] = "Error";
})(Status || (Status = {}));
export class Donate {
    constructor() {
        this.status = Status.Idle;
        this.isOpen = false;
        this.addressChunks = [];
        /**
         * The theme of the button. Use light when using against dark background.
         */
        this.theme = 'neutral';
        /**
         * The network you want to use. Can be 'main' or 'test'.
         */
        this.network = 'main';
        /**
         * The text to display on the button. Default: "Donate NIM"
         */
        this.text = 'Donate NIM';
        /**
         * The text to display once payment is done. Default: "Thank you for the donation!"
         */
        this.thankyou = 'Thank you for the donation!';
    }
    onRecipientChanged() {
        this.breakAddressIntoChunks();
        this.generateIdenticon();
    }
    componentWillLoad() {
        Iqons.svgPath = getAssetPath('./identicons/iqons.min.svg');
        this.breakAddressIntoChunks();
        this.generateIdenticon();
    }
    breakAddressIntoChunks() {
        const addressWithoutSpaces = this.recipient.replace(/\s/g, '');
        this.addressChunks = addressWithoutSpaces.match(/.{1,4}/g);
    }
    generateIdenticon() {
        Iqons.toDataUrl(this.recipient).then(identicon => {
            this.identicon = identicon;
        });
    }
    open() {
        this.isOpen = true;
    }
    close() {
        this.isOpen = false;
    }
    async donate(event) {
        event.preventDefault();
        this.status = Status.Progress;
        let options = {
            appName: this.text,
            recipient: this.recipient,
            value: this.amount * 100000,
        };
        if (this.logoUrl) {
            options.shopLogoUrl = this.logoUrl;
        }
        if (this.fee) {
            options.fee = this.fee;
        }
        if (this.message) {
            options.extraData = this.message.trim();
        }
        const hubApi = new HubApi(HubUrl[this.network]);
        try {
            await hubApi.checkout(options);
            this.status = Status.Success;
            setTimeout(() => {
                this.status = Status.Idle;
                this.close();
            }, 2000);
        }
        catch (error) {
            this.status = Status.Error;
            setTimeout(() => {
                this.status = Status.Idle;
            }, 2000);
        }
    }
    updateAmount(event) {
        this.amount = parseFloat(event.target.value);
    }
    updateMessage(event) {
        this.message = event.target.value;
    }
    canConfirm() {
        return this.amount > 0;
    }
    render() {
        return (h("div", { class: "donate" },
            h("button", { class: 'nim-button nim-theme-' + this.theme, disabled: this.status === Status.Progress, type: "button", onClick: _ => this.open() },
                h(Hexagon, { class: "nim-icon nim-logo" }),
                h("div", null,
                    h("span", null, this.text))),
            h("div", { class: { overlay: true, 'is-open': this.isOpen } },
                h("div", { class: "dialog" },
                    h("button", { onClick: _ => this.close(), type: "button", class: "close" },
                        h(Close, { class: "nim-icon" })),
                    h("form", { onSubmit: e => this.donate(e) },
                        h("h2", null, "Donate"),
                        h("div", { class: "recipient" },
                            h("div", { class: "identicon" },
                                h("img", { src: this.identicon })),
                            h("div", { class: "address" }, this.addressChunks.map(chunk => (h("span", { class: "chunk" }, chunk))))),
                        h("div", { class: "amount" },
                            h("input", { onInput: e => this.updateAmount(e), type: "text", placeholder: "0.00" }),
                            h("span", { class: "ticker" }, "NIM")),
                        h("div", { class: "message" },
                            h("input", { onInput: e => this.updateMessage(e), type: "text", maxLength: 64, placeholder: "Add a message..." })),
                        h("div", { class: "confirm" },
                            h("button", { disabled: !this.canConfirm(), class: "nim-button nim-theme-light-blue", type: "submit" }, "Confirm"))),
                    h("div", { class: {
                            flash: true,
                            show: this.status !== Status.Idle,
                        } },
                        h("p", null, "Transaction in progress...")),
                    h("div", { class: {
                            'flash success': true,
                            show: this.status === Status.Success,
                        } },
                        h("p", null, this.thankyou)),
                    h("div", { class: {
                            'flash error': true,
                            show: this.status === Status.Error,
                        } },
                        h("p", null, "Transaction failed. Please try again."))))));
    }
    static get is() { return "nim-donate"; }
    static get encapsulation() { return "shadow"; }
    static get originalStyleUrls() { return {
        "$": ["donate.css"]
    }; }
    static get styleUrls() { return {
        "$": ["donate.css"]
    }; }
    static get assetsDirs() { return ["identicons"]; }
    static get properties() { return {
        "theme": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "Theme",
                "resolved": "\"blue\" | \"gold\" | \"green\" | \"light-blue\" | \"neutral\" | \"orange\" | \"red\"",
                "references": {
                    "Theme": {
                        "location": "import",
                        "path": "../../utils/common"
                    }
                }
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The theme of the button. Use light when using against dark background."
            },
            "attribute": "theme",
            "reflect": false,
            "defaultValue": "'neutral'"
        },
        "network": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "Network",
                "resolved": "\"main\" | \"test\"",
                "references": {
                    "Network": {
                        "location": "import",
                        "path": "../../utils/common"
                    }
                }
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The network you want to use. Can be 'main' or 'test'."
            },
            "attribute": "network",
            "reflect": false,
            "defaultValue": "'main'"
        },
        "recipient": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The human-readable address of the recipient (your shop/app)."
            },
            "attribute": "recipient",
            "reflect": false
        },
        "logoUrl": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "An image URL. Must be on the same origin as the request is sent from.\nShould be square and at least 146x146 px."
            },
            "attribute": "logo-url",
            "reflect": false
        },
        "fee": {
            "type": "number",
            "mutable": false,
            "complexType": {
                "original": "number",
                "resolved": "number",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "Transaction fee in luna. Default: 0"
            },
            "attribute": "fee",
            "reflect": false
        },
        "text": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The text to display on the button. Default: \"Donate NIM\""
            },
            "attribute": "text",
            "reflect": false,
            "defaultValue": "'Donate NIM'"
        },
        "thankyou": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The text to display once payment is done. Default: \"Thank you for the donation!\""
            },
            "attribute": "thankyou",
            "reflect": false,
            "defaultValue": "'Thank you for the donation!'"
        }
    }; }
    static get states() { return {
        "status": {},
        "isOpen": {},
        "addressChunks": {},
        "identicon": {},
        "amount": {},
        "message": {}
    }; }
    static get watchers() { return [{
            "propName": "recipient",
            "methodName": "onRecipientChanged"
        }]; }
}
